/*
 * bufcomm.cc
 *
 *  Created on: 06.05.2012
 *      Author: erlkoenig
 */

#include "stm32f4xx.h"
#include <string.h>
#include <algorithm>

#include "bufcomm.h"

RingBuffer::RingBuffer(len_t size, uint8_t* mem) : m_size (size), m_fill (0), m_wptr (0), m_rptr (0), m_buffer (mem) {
	TIM_SetCompare1(TIM4,1000);
}

void RingBuffer::write (const uint8_t* data, RingBuffer::len_t len) {
	len_t t = std::min(len,(uint16_t)(m_size-m_wptr));
	memcpy (&(m_buffer [m_wptr]), data, t);
	if (t < len) {
		memcpy (m_buffer, &(data[t]), len-t);
		m_wptr = len-t;
	} else {
		m_wptr += len;
		if (m_wptr == m_size) m_wptr = 0;
	}
	m_fill += len;
}

void RingBuffer::read (uint8_t* buffer, RingBuffer::len_t len) {
	len_t t = std::min(len,(len_t)(m_size-m_rptr));
	memcpy (buffer, &(m_buffer [m_rptr]), t);
	if (t < len) {
		memcpy (&(buffer[t]), m_buffer, len-t);
		m_rptr = len-t;
	} else {
		m_rptr += len;
		if (m_rptr == m_size) m_rptr = 0;
	}
	m_fill -= len;
}

void RingBuffer::write_byte (uint8_t byte) {
	m_buffer [m_wptr] = byte; m_wptr++;
	if (m_wptr == m_size) m_wptr = 0;
	m_fill++;
}

uint8_t RingBuffer::read_byte () {
	uint8_t r = m_buffer [m_rptr];
	m_rptr++;
	if (m_rptr == m_size) m_rptr = 0;
	m_fill--;
	return r;
}

void RingBuffer::clear () {
	m_fill = m_rptr = m_wptr = 0;
}

BufferedComm::BufferedComm () : m_timeout (false), m_istate (0), m_ostate (0),
	m_peer_size (0), m_sent_size (0), m_rec_size(0), m_send_size (0),
	m_in (sizeof (m_buf_in), m_buf_in), m_out (sizeof (m_buf_out), m_buf_out) {
}

void BufferedComm::on_rx (uint8_t data) {
	if (m_istate == 0) {
		if (data == 2) {
			set_timer (false); m_timeout = false;
			m_istate = 2;
			if (m_ostate == 0) {
				m_ostate = 1;
			}
			on_tx_idle ();
		} else if (data == 1 && m_ostate == 0) {
			m_ostate = 1;
			set_timer (false); m_timeout = false;
			on_tx_idle ();
		}
	} else if (m_istate == 2) {
		if (data == 3) {
			m_istate = 3;
		} else if (data == 4) {
			m_istate = 5;
		} else {
			// ! Error
		}
	} else if (m_istate == 3) {
		m_peer_size += data;
		m_istate++;
	} else if (m_istate == 4) {
		m_peer_size += ((RingBuffer::len_t) (data)) << 8;
		m_istate = 2;
		on_tx_idle ();
	} else if (m_istate == 5) {
		m_rec_size = data; m_istate++;
	} else if (m_istate == 6) {
		m_in.write_byte (data);
		m_sent_size--;
		m_rec_size--;
		if (m_rec_size == 0) {
			m_istate = 2;
			on_rec ();
		}
	}
}

void BufferedComm::on_tx_idle () {
	if (!tx_idle ()) { set_idle (true); return; }

	if (m_ostate == 1) {
		write (2); m_ostate = 2;
	} else if (m_istate == 0 && m_ostate == 0) {
		if (!m_timeout) { set_timer (true); set_idle (false); return; }
		write (1);
		m_timeout = false;
	} else if (m_istate >= 2) {
		if (m_ostate == 2 && m_in.available () > m_sent_size) {
			m_ostate = 3;
			write (3);
		} else if (m_ostate == 2 && m_out.fill () > 0 && m_peer_size > 0) {
			write (4);
			m_ostate = 5;
		} else if (m_ostate == 3) {
			m_send_size = m_in.available () - ((RingBuffer::len_t) (m_sent_size));
			m_sent_size += m_send_size;
			write (m_send_size & 0xFF);
			m_ostate++;
		} else if (m_ostate == 4) {
			write (m_send_size >> 8);
			m_ostate = 2;
		} else if (m_ostate == 5) {
			m_send_size = std::min (m_peer_size, m_out.fill ());
			if (m_send_size > 255) m_send_size = 255;
			m_peer_size -= m_send_size;
			write (m_send_size);
			m_ostate++;
		} else if (m_ostate == 6) {
			write (m_out.read_byte ());
			m_send_size--;
			if (m_send_size == 0) {
				m_ostate = 2;
				on_sent ();
			}
		} else {
			set_idle (false); return;
		}
	} else {
		set_idle (false); return;
	}
	set_idle (true);
}

void BufferedComm::on_timer () {
	if (m_istate == 0) { m_timeout = true; set_timer (false); set_idle (true); }
}

void BufferedComm::reset () {
	m_timeout = false; m_istate = 0; m_ostate = 0;
	m_peer_size = 0; m_sent_size = 0; m_rec_size = 0; m_send_size = 0;
	m_in.clear (); m_out.clear ();
}
